/* Ambiguous Frequent Itemset Minor */
/* 2007/9/30 Takeaki Uno,   e-mail:uno@nii.jp, 
    homepage:   http://research.nii.ac.jp/~uno/index.html  */
/* This program is available for only academic use, basically.
   Anyone can modify this program, but he/she has to write down 
    the change of the modification on the top of the source code.
   Neither contact nor appointment to Takeaki Uno is needed.
   If one wants to re-distribute this code, please
    refer the newest code, and show the link to homepage of 
    Takeaki Uno, to notify the news about LCM for the users. */

#ifndef _afim_c_
#define _afim_c_

#define WEIGHT_DOUBLE

#include"alist.c"
#include"trsact.c"
#include"problem.c"

#define AFIM_IGNORE_SMALL 16
#define AFIM_TRSACT_ID_MAXIMAL 128

void AFIM_error (){
  ERROR_MES = "command explanation";
  print_err ("AFIM: [FMfIq] [options] input-filename support density-th [output-filename]\n\
%%:show progress, _:no message, +:write solutions in append mode\n\
F:ambiguous frequent itemset mining, M:maximal ambiguous frequent itemset mining\n\
f,Q:output frequency following/preceding to itemsets\n\
I:output ID's of transactions including each pattern, m: output ID's maximaly\n\
i:ignore item s.t. ambiguous frequency of its singleton is less than support\n\
t:transpose the database so that i-th transaction will be item i\n\
[options]\n\
-K [num]:output [num] most frequent itemsets\n\
-l [num]:output itemsets with size at least [num]\n\
-u [num]:output itemsets with size at most [num]\n\
-U [num]:upper bound for support(maximum support)\n\
-w [filename]:read weights of transactions from the file\n\
-# [num]:stop after outputting [num] solutions\n\
-, [char]:give the separator of the numbers in the output\n\
-Q [filename]:replace the output numbers according to the permutation table given by [filename]\n\
# the 1st letter of input-filename cannot be '-'.\n\
# if the output file name is -, the solutions will be output to standard output.\n");
  EXIT;
}

/***********************************************************************/
/*  read parameters given by command line  */
/***********************************************************************/
void AFIM_read_param (int argc, char *argv[], PROBLEM *PP){
  int c=1;
  ITEMSET *II = &PP->II;

  if ( argc<c+4 ){ AFIM_error (); return; }
  
  if ( !strchr (argv[c], '_') ){ II->flag |= SHOW_MESSAGE; PP->TT.flag |= SHOW_MESSAGE; }
  if ( strchr (argv[c], '%') ) II->flag |= SHOW_PROGRESS;
  if ( strchr (argv[c], '+') ) II->flag |= ITEMSET_APPEND;
  if ( strchr (argv[c], 'f') ) II->flag |= ITEMSET_FREQ;
  if ( strchr (argv[c], 'Q') ) II->flag |= ITEMSET_PRE_FREQ;
  if ( strchr (argv[c], 'M') ) PP->problem = PROBLEM_MAXIMAL;
  else if ( strchr (argv[c], 'F') ) PP->problem = PROBLEM_FREQSET;
  else error ("F or M command has to be specified.", EXIT);
  if ( strchr (argv[c], 'I') ) II->flag |= ITEMSET_TRSACT_ID;
  if ( strchr (argv[c], 'm') ) PP->problem |= AFIM_TRSACT_ID_MAXIMAL;
  if ( strchr (argv[c], 'i') ) II->flag |= AFIM_IGNORE_SMALL;
  if ( strchr (argv[c], 't') ) PP->TT.flag |= LOAD_TPOSE;
  c++;

  while ( argv[c][0] == '-' ){
    switch (argv[c][1]){
      case 'K': II->topk.end = atoi (argv[c+1]);
      break; case 'l': II->lb = atoi (argv[c+1]);
      break; case 'u': II->ub = atoi(argv[c+1]);
      break; case 'U': II->frq_ub = (WEIGHT)atof(argv[c+1]);
      break; case 'w': PP->TT.wfname = argv[c+1];
      break; case '#': II->max_solutions = atoi(argv[c+1]);
      break; case ',': II->separator = argv[c+1][0];
      break; case 'Q': PP->outperm_fname = argv[c+1];
      break; default: goto NEXT;
    }
    c += 2;
    if ( argc < c+3 ){ AFIM_error (); return; }
  }

  NEXT:;
  PP->TT.fname = argv[c];
  PP->dense = atof(argv[c+2]);
  if ( II->topk.end==0 ) II->frq_lb = (WEIGHT)atof(argv[c+1]);
  if ( argc>c+3 ){
    II->flag |= 1;
    PP->output_fname = argv[c+3];
  }
}


/*************************************************************************/
/* AFIM main routine */
/*************************************************************************/
void AFIM (PROBLEM *PP){
  VEC_ID t, tt, i, remain;
  int flag;
  QUEUE_INT e, th_h, h, *x, *y;
  size_t cnt, k;
  WEIGHT frq;
  ITEMSET *II = &PP->II;
  TRSACT *TT = &PP->TT;
  QUEUE CHD;

  II->iters++;
  frq = II->frq;
//printf ("## %d\n", II->iters);

//MQUE_FLOOP (II->itemset, x) printf ("%d ", II->perm[*x]);; printf (": itemset\n"); 
//printf ("frq = %f\n", II->frq);
     // min.sup ȉłő̐؂ th_h (k*(P))vZ

  for (th_h=II->itemset.t,k=0 ; k+PP->occ.num[th_h]<II->frq_lb ; th_h--)
      k += PP->occ.num[th_h];
  
     // ؂ th_h ܂ IJXU蕪(JEĝ)
  TT->jump.t = 0;
  for (cnt=0,h=II->itemset.t ; h>th_h ; h--){
    MALIST_DO_FORWARD (PP->occ, h, e)
        TRSACT_delivery_iter (TT, &TT->jump, PP->occ_w, PP->occ_pw, e, TT->T.clms);
    cnt += PP->occ.num[h] * h;
  }
  
      // P+e  min.sup ȏɂȂ\̂̂̂݁AsbNAbvBjumpɎc(łȂ̂́AOccNA)
      // ɂāAmin.sup ܂łɂƂKvvZ
//  remain = (int)ceil(PP->dense*(II->frq_lb*(II->itemset.t+1))) - cnt - (II->frq_lb - k) * th_h;

  remain = (VEC_ID)(PP->dense*(II->frq_lb*(II->itemset.t+1)));
  if ( PP->dense*(II->frq_lb*(II->itemset.t+1)) > remain ) remain++; // take the ceiling
  remain = (VEC_ID)(remain - cnt - (II->frq_lb - k) * th_h);

//  printf ("cnt=%d, remain %d %d \n", cnt, remain, (int)ceil(PP->dense*(II->frq_lb*(II->itemset.t+1))));

  MQUE_FLOOP (II->itemset, x) PP->itemflag[*x] = INTHUGE; // ǉς݃ACe폜
  PP->itemcand.t = 0;
  if ( remain <= 0 ){
    FLOOP (e, 0, TT->T.clms)
        if ( PP->itemflag[e] != INTHUGE ) QUE_INS (PP->itemcand, e);
  } else {
    MQUE_FLOOP (TT->jump, x){
//printf ("e=%d, %d %d\n", e, remain-QQ[e].t, II->frq_lb - k);
      if ( remain-TT->OQ[*x].end > II->frq_lb - k )
          PP->itemflag[*x] = INTHUGE;  // poɂȂ錩݂Ȃ
       // frequency ݂̌܂ł̘aۑ
    }

      // ؂ th_h U蕪(JEĝ)
    MALIST_DO_FORWARD (PP->occ, th_h, e)
        TRSACT_delivery_iter (TT, &TT->jump, PP->occ_w, PP->occ_pw, e, TT->T.clms);
    MQUE_FLOOP (TT->jump, x){
//      printf ("##e=%d, occ.t=%d, remain=%d\n", e, QQ[*x].t, remain);
      if ( PP->itemflag[*x] != INTHUGE && TT->OQ[*x].end >= remain ){
           // P+e ܂poW
        QUE_INS (PP->itemcand, *x);
      }
      PP->itemflag[*x] = 0;
    }
  }

       // OQ.end ̃NA
  MQUE_FLOOP (TT->jump, x) TT->OQ[*x].end = 0;
  
      // ǉς݃ACeɊւtO
  MQUE_FLOOP (II->itemset, x) PP->itemflag[*x] = 0;
      //  cand -> CHD ɎqǂRs[
  QUEUE_alloc (&CHD, PP->itemcand.t);
  MQUE_FLOOP (PP->itemcand, x) QUE_INS (CHD, *x);
      // eɂāAevZ

//printf ("CHD: ");QUEUE_print__ (&CHD);  
  MQUE_FLOOP (CHD, x){

//printf ("%d    ", ee); QUEUE_print ( &II->itemset);
      // occ, itemset ̍XV
//    for (i=0 ; i<TT->OQ[ee].t ; i++){
//        t = TT->OQ[ee].v[i];

    MQUE_FLOOP (TT->OQ[*x], y) MALIST_mv (&PP->occ, PP->occ.list[*y]+1, *y, 0);
    QUE_INS (II->itemset, *x);
    flag = 1;  // qǂǂ̃tO
    if ( II->itemset.t== 1 ) goto LOOP_CHILD;  // 傫0->1 ȂKq
    
        // k*(P+e) ̌vZ
//printf ("%p : ", x); FLOOP (cnt, 0, II->itemset.t+1) printf ("%d, ", PP->occ.num[cnt]); printf (" item=%d\n", *x);
//MALIST_print (&PP->occ, II->itemset.t);
//QUEUE_print__  (&TT->OQ[*x]);
    for (cnt=0,II->frq=0,h=II->itemset.t ; h>=0 ; h--){
      if ( PP->dense*((II->frq+PP->occ.num[h])*(II->itemset.t)) > cnt+PP->occ.num[h] * h ) break;
      II->frq += PP->occ.num[h];
      cnt += PP->occ.num[h] * h;
    }

// printf ("k* = %d, t=%d cnt=%d\n", h, II->frq, cnt);
    if ( II->frq > frq ) goto LOOP_END;  // pox傫ȂqǂłȂ

       // cov(P+e)  AmbiOcc ̌vZ
    TT->jump.t = 0;
    if ( II->frq == TT->T.t ) h = -1;
    for (i=II->itemset.t ; i>h ; i--)
        MALIST_DO_FORWARD (PP->occ, i, e)
            TRSACT_delivery_iter (TT, &TT->jump, PP->occ_w, PP->occ_pw, e, TT->T.clms);
    tt = (VEC_ID)((((double)cnt)-PP->dense*(II->itemset.t)*II->frq) / (PP->dense*II->itemset.t-h));  // tt = Occ_h ̒ AmbiOcc ɓgUNV
    if ( II->frq < TT->T.t && tt > 0 ){
      MALIST_k_partition (&PP->occ, h, tt, 0, 0);
      II->frq += tt;

//    printf ("tt = %d,   %f %f %f\n", tt, (double)cnt, PP->dense*(II->itemset.t)*t, PP->dense*II->itemset.t-h );
  //  cnt+hx == PP->dense*(itemset.t)*(t+x)
  //  (cnt-PP->dense*(itemset.t+1)*t) / PP->dense*itemset.t-h) == x
  //  (cnt+kx)/(l+x)t >= th


      for (i=0,t=MALIST_HEAD(PP->occ,h) ; i<tt ; i++,t=PP->occ.nxt[t]){
        MQUE_FLOOP (TT->T.v[t], y){
          if ( TT->OQ[*y].t == 0 ) QUE_INS (TT->jump, *y);
          TT->OQ[*y].end++;
        }
      }
    }
    
      // e*(P+e) vZĐẽ`FbN
    MQUE_FLOOP (II->itemset, y){
//printf ("e=%d, occ.t=%d\n", e, QQ[e].t);
      if ( TT->OQ[*y].end < TT->OQ[*x].end || ( TT->OQ[*y].end==TT->OQ[*x].end && *y<*x) ) flag=0;
    }  // e*  e ł͂Ȃ̂ŁAqǂł͂Ȃ

       //  occ ̃NA
    MQUE_FLOOP (TT->jump, y) TT->OQ[*y].end = 0;

    LOOP_CHILD:;
    if ( flag ) AFIM (PP);  // qǂȂċAĂяo

    LOOP_END:;
         // occ, itemset ̉
//    for (i=0 ; i<PP->occ_t[*x] ; i++){
//      t = ((QUEUE *)TT->Occ.h)[*x].v[i];
    MQUE_FLOOP (TT->OQ[*x], y) MALIST_mv (&PP->occ, PP->occ.list[*y]-1, *y, 0);
    II->itemset.t--;
  }
  
  
       // ̏o
  II->frq = frq;
  if ( (PP->problem&PROBLEM_MAXIMAL)==0 || CHD.t==0 ){
    PP->veccand.t = 0;
    if ( II->flag & ITEMSET_TRSACT_ID ){
      for (cnt=0,h=II->itemset.t ; h>0 ; h--){
        if ( cnt>0 && ((double)cnt+h) / (II->itemset.t*(PP->veccand.t+1)) < PP->dense ) break; // tight case; not dense even if we add one transaction
        cnt += PP->occ.num[h] * h;
        if ( !(PP->problem&AFIM_TRSACT_ID_MAXIMAL) && cnt>0 && ((double)cnt) / (II->itemset.t*(PP->veccand.t+PP->occ.num[h])) < PP->dense ) break;  // trasact set is minimal
        MALIST_DO_FORWARD (PP->occ, h, e) QUE_INS (PP->veccand, e);
        if ( cnt>0 && ((double)cnt) / (II->itemset.t*PP->veccand.t) < PP->dense ) break;
      }
    }
    ITEMSET_output_itemset (II, &PP->veccand, 0);
  }
  QUEUE_end (&CHD);
}



/*************************************************************************/
/* main function of AFIM */
/*************************************************************************/
int AFIM_main (int argc, char *argv[]){
  PROBLEM PP;
  ITEMSET *II = &PP.II;
  TRSACT *TT = &PP.TT;
  QUEUE_ID i;

  PROBLEM_init ( &PP);
  AFIM_read_param (argc, argv, &PP);
  if ( ERROR_MES ) return (1);
//  TT->flag = LOAD_RM_DUP +LOAD_PERM +TRSACT_FRQSORT +LOAD_DECSORT +TRSACT_ALLOC_OCC;
  TT->flag |= LOAD_RM_DUP +TRSACT_FRQSORT +LOAD_DECSORT +TRSACT_ALLOC_OCC;
  TT->th = (II->flag&AFIM_IGNORE_SMALL)? (double)II->frq*PP.dense :(double)1;

  PROBLEM_load (&PP);
  if (PP.II.flag & SHOW_MESSAGE ) fprintf (stderr, "density_threashold= %f\n", PP.dense);
  PROBLEM_alloc (&PP, TT->T.clms, TT->T.t, TT->T.t, TT->perm, PROBLEM_ITEMFLAG +PROBLEM_ITEMCAND +((TT->flag&TRSACT_NEGATIVE)?PROBLEM_OCC_PW: PROBLEM_OCC_W) +PROBLEM_VECCAND );
  QUEUE_delivery (TT->OQ, NULL, NULL, TT->T.v, &TT->OQ[TT->T.clms], TT->T.t, TT->T.clms);
  II->frq = TT->total_w_org;
  TT->perm = NULL;
  II->X = (void *)TT;
// delivery
  MALIST_alloc (&PP.occ, TT->T.t+2, TT->T.t+2); // element=>item, list => trsact
  FLOOP (i, 0, TT->T.clms) TT->OQ[i].end = 0;
  FLOOP (i, 0, TT->T.t) MALIST_ins_tail (&PP.occ, 0, i, 0);
  if ( !ERROR_MES ){
    AFIM (&PP);
    ITEMSET_last_output (&PP.II);
  }
  PROBLEM_end (&PP);
  return (ERROR_MES?1:0);
}

/*******************************************************************************/
#ifndef _NO_MAIN_
#define _NO_MAIN_
int main (int argc, char *argv[]){
  return (AFIM_main (argc, argv));
}
#endif
/*******************************************************************************/

#endif
